/*---------------------------------------------------------------------------
	Project:	    WL33_NUCLEO_UART

	File Name:	    chat.c

	Author:		    MartinA

	Description:	This code contains the chat application. It gathers key until the CR
					is hit to send a packet. Two internal operations are exit and set the
					destination.

					This program is free software: you can redistribute it and/or modify
					it under the terms of the GNU General Public License as published by
					the Free Software Foundation, either version 2 of the License, or
					(at your option) any later version, provided this copyright notice
					is included.

				    Copyright (c) Alberta Digital Radio Communications Society
				    All rights reserved.

	Revision History:

---------------------------------------------------------------------------*/
#include <stdint.h>
#include <string.h>
#include <malloc.h>

#include "setup.h"
#include "frame.h"
#include "types.h"
#include "usart.h"
#include "streambuffer.h"
#include "dataq.h"

#define MAX_KEY		140				// max keys in a buffer
#define	MAX_DEST	20				// max chars in dest callsign
#define	BROADCAST	"FFFF"			// broadcast address

static char keyBuffer[MAX_KEY];		// buffer for keystrokes
uint8_t keyPos;						// current position

#define	KEY_EOL			0x0D		// carriage return
#define KEY_RPT			0x12		// change repeat status
#define	KEY_EXIT		0x1A		// exit key
#define	KEY_ESC			0x1B		// escape key
#define KEY_DEL			0x7F		// delete key
#define	KEY_DUMP		0x04		// toggle dump mode

char dest_call[MAX_DEST];			// broadcast destination
char *dp = dest_call;				// pointer to dest call characters
char *entCall = "Enter Destination Callsign";
char *rptMode[] = {
		"Repeat mode->off",
		"Repeat mode->on"
};

char *dumpStrings[] = {
		"Dump mode->off",
		"Dump mode->on"
};

char *welcome = "Welcome to chat. ESC to set destination, CTRL/R to toggle repeat, CTRL/Z to exit";

BOOL destEnt = FALSE;				// not entering destination
BOOL repeat = TRUE;					// repeating by default
BOOL deleteMode = FALSE;			// delete mode
BOOL welcomed = FALSE;				// welcome mat is out
BOOL dumpMode = FALSE;

FRAME_QUEUE chatQueue;				// queue for inbound frames

// fwd refs
void sendLine(char *buffer, int len);
void PrintFrame(uint8_t *FrameBytes);

// init entry
void Chat_Task_init(void)
{
	strcpy(dest_call, BROADCAST);
	destEnt = FALSE;
	chatQueue.q_forw = &chatQueue;
	chatQueue.q_back = &chatQueue;
}

void Chat_Task_welcome(void)
{
	// start with welcome message
	USART_Print_string("%s\r\n", welcome);
	USART_Print_string("%s; ", rptMode[repeat]);

	if(!strcmp(dest_call, BROADCAST))
		USART_Print_string("%s; Destination callsign->(Broadcast)\r\n\n");
	else
		USART_Print_string("%s; Destination callsign->%s\r\n\n", dest_call);

	welcomed = TRUE;
}

/*
 * place a frame on the queue frame content is copied
 * to alloc'd memory
 */
BOOL EnqueChatFrame(void *Frame, int length)
{
	uint8_t *FrameData;
	FRAME_QUEUE *f = NULL;

	if((f=(FRAME_QUEUE *)malloc(sizeof(FRAME_QUEUE))) == NULL)
		return FALSE;

	// copy the frame data in..
	if((FrameData = malloc(length)) == NULL)	{
		free(f);
		return FALSE;
	}
	memcpy(FrameData, Frame, length);

	f->frame = FrameData;
	f->length = length;
	insque((QUEUE_ELEM *)f, (QUEUE_ELEM *)chatQueue.q_back);

	return TRUE;
}

/*
 * chat task main entry
 * return: 	TRUE if return to main menu is required
 * 			FALSE: more to go
 */
BOOL Chat_Task_exec(void)
{
	if(!welcomed)
		Chat_Task_welcome();

	char c;
	int nBytesinBuff;

	// process any inbound frames first..
	if((keyPos == 0) && chatQueue.q_forw != &chatQueue)	{
		FRAME_QUEUE *f = chatQueue.q_forw;
		remque((struct qelem *)f);
		PrintFrame(f->frame);
		free(f->frame);
		free(f);
	}

	if((nBytesinBuff=databuffer_bytesInBuffer()) == 0)
		return FALSE;

	for(int i=0;i<nBytesinBuff;i++)		{

		c=databuffer_get(0);

		if(deleteMode)	{
			if(c != KEY_DEL)	{
				USART_Print_string("\\%c", c);
				if(keyPos < MAX_KEY)
					keyBuffer[keyPos++] = c;
				deleteMode = FALSE;
			} else {
				if(keyPos > 0)	{
					USART_Print_string("%c", keyBuffer[--keyPos]);
				} else {
					USART_Print_string("\\\r\n");
					deleteMode = FALSE;
				}
			}
			continue;
		} else {
		// processing a key

			switch (c)	{

			// CTRL/R: change repeat flag
			case KEY_RPT:
				repeat = repeat ? FALSE : TRUE;
				char *r = rptMode[repeat];
				USART_Print_string("%s\r\n",r);
				break;

			// CTRL/D: change dump mode
			case KEY_DUMP:
				dumpMode = dumpMode ? FALSE : TRUE;
				char *d = dumpStrings[dumpMode];
				USART_Print_string("%s\r\n",d);
				break;

			// escape key: get a destination call sign
			case KEY_ESC:
				if(keyPos == 0)	{
					USART_Print_string("%s->", entCall);
					dp = dest_call;
					destEnt = TRUE;
				}
				break;

				// EOL key: sent the packet
			case KEY_EOL:
				USART_Print_string("\r\n");
				if(destEnt)	{
					int cpyLen = keyPos > MAX_CALL ? MAX_CALL : keyPos;
					strncpy(dest_call, keyBuffer, cpyLen);
					destEnt = FALSE;
				} else {
					keyBuffer[keyPos++] = '\0';
					sendLine(keyBuffer, keyPos);
				}
				keyPos = 0;
				break;

			case KEY_EXIT:
				welcomed = FALSE;
				return TRUE;

			case KEY_DEL:
				USART_Print_string("\\%c", keyBuffer[--keyPos]);
				deleteMode = TRUE;
				break;

			default:
				USART_Send_Char(c);
				if(keyPos < MAX_KEY)
					keyBuffer[keyPos++] = c;
				break;
			}
		}
	}
	return FALSE;
}

// send a line of text
void sendLine(char *buffer, int len)
{
	SendTextFrame(setup_memory.params.setup_data.stnCall, TEXT_TYPE, dest_call, TEXT_TYPE, buffer, len, repeat);
}

/*
 * Print a received frame on the console
 */
void PrintFrame(uint8_t *FrameBytes)
{
	char printBuf[250];
	char decCall[100];
	uint16_t port;
	uint16_t dataLen;

	// aid in decoding callsigns
	union {
		IP400_CALL	callsign;
		uint8_t bucketofBytes[IP_400_CALL_SIZE];
	} callBytes;

	union {
		IP400_FLAGS flags;
		uint8_t bucketofFlags[IP_400_FLAG_SIZE];
	} flagBytes;


	// dump mode for header debugging
	if(dumpMode)	{
	      /* print the received data */
	      USART_Print_string("RX - Data received: [ ");

	      for(uint8_t i=0; i<sizeof(IP400_FRAME); i++)
	        USART_Print_string("%02x ", FrameBytes[i]);

	      USART_Print_string("]\r\n");
	      return;
	}


	//decode the 'to' callsign
	memcpy(callBytes.bucketofBytes, FrameBytes, IP_400_CALL_SIZE);
	FrameBytes += IP_400_CALL_SIZE;
	callDecode(&callBytes.callsign, decCall, &port);
	USART_Print_string("%s(%d)>", decCall, port);

	// now the 'from'
	memcpy(callBytes.bucketofBytes, FrameBytes, IP_400_CALL_SIZE);
	FrameBytes += IP_400_CALL_SIZE;
	if((callBytes.callsign.callbytes.bytes[0] == BROADCAST_ADDR)	 &&
		(callBytes.callsign.callbytes.bytes[1] == BROADCAST_ADDR)) {
		USART_Print_string("BROADCAST(%d)", callBytes.callsign.port);
	} else {
		callDecode(&callBytes.callsign, decCall, &port);
		USART_Print_string("%s(%d)", decCall, port);
	}

	// get the flags
	memcpy(flagBytes.bucketofFlags, FrameBytes, IP_400_FLAG_SIZE);
	FrameBytes += IP_400_FLAG_SIZE;
	USART_Print_string("[%d]:", flagBytes.flags.hop_count);

	// now dump the data
	memcpy(&dataLen, FrameBytes, sizeof(uint16_t));
	FrameBytes += sizeof(uint16_t);
	memcpy(printBuf, FrameBytes, dataLen);
	printBuf[dataLen] = '\0';
	USART_Print_string("%s\r\n", printBuf);

}
